Package: oskeyring
Title: Raw System Credential Store Access from R
Version: 0.1.7
Authors@R: c(
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Aims to support all features of the system credential store,
    including non-portable ones. Supports 'Keychain' on 'macOS', and
    'Credential Manager' on 'Windows'. See the 'keyring' package if you
    need a portable 'API'.
License: MIT + file LICENSE
URL: https://github.com/r-lib/oskeyring#readme,
        https://r-lib.github.io/oskeyring/
BugReports: https://github.com/r-lib/oskeyring/issues
Depends: R (>= 3.6)
Suggests: askpass, covr, testthat (>= 3.0.0), withr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-04-30
Encoding: UTF-8
RoxygenNote: 7.2.3
NeedsCompilation: yes
Packaged: 2026-04-08 08:35:18 UTC; gaborcsardi
Author: Gábor Csárdi [aut, cre],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-08 09:00:09 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-04-08 10:47:06 UTC; unix
Archs: oskeyring.so.dSYM

Package: oskeyring
Title: Raw System Credential Store Access from R
Version: 0.1.7
Authors@R: c(
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Aims to support all features of the system credential store,
    including non-portable ones. Supports 'Keychain' on 'macOS', and
    'Credential Manager' on 'Windows'. See the 'keyring' package if you
    need a portable 'API'.
License: MIT + file LICENSE
URL: https://github.com/r-lib/oskeyring#readme,
        https://r-lib.github.io/oskeyring/
BugReports: https://github.com/r-lib/oskeyring/issues
Depends: R (>= 3.6)
Suggests: askpass, covr, testthat (>= 3.0.0), withr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-04-30
Encoding: UTF-8
RoxygenNote: 7.2.3
NeedsCompilation: yes
Packaged: 2026-04-08 08:35:18 UTC; gaborcsardi
Author: Gábor Csárdi [aut, cre],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-08 09:00:09 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 20:59:23 UTC; unix
Archs: oskeyring.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(format,oskeyring_macos_item)
S3method(format,oskeyring_windows_item)
S3method(print,oskeyring_macos_item)
S3method(print,oskeyring_windows_item)
export(macos_item)
export(macos_item_add)
export(macos_item_attr)
export(macos_item_classes)
export(macos_item_delete)
export(macos_item_match_options)
export(macos_item_search)
export(macos_item_update)
export(macos_keychain_create)
export(macos_keychain_delete)
export(macos_keychain_is_locked)
export(macos_keychain_list)
export(macos_keychain_lock)
export(macos_keychain_unlock)
export(windows_item)
export(windows_item_delete)
export(windows_item_enumerate)
export(windows_item_read)
export(windows_item_types)
export(windows_item_write)
useDynLib(oskeyring, .registration = TRUE)

# oskeyring

> Raw System Credential Store Access from R

<!-- badges: start -->

[![R-CMD-check](https://github.com/r-lib/oskeyring/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/oskeyring/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/oskeyring/graph/badge.svg)](https://app.codecov.io/gh/r-lib/oskeyring)
<!-- badges: end -->

## Features

- Windows and macOS support. Read, write, list and search the system
  credential store.
- Generic credentials, domain passwords, domain certificates on Windows.
- Generic passwords and internet passwords on macOS.
- Multiple keychains on macOS.

## Related

- The keyring R package provides a portable system keyring API for all
  platforms, and also supports multiple backends:
  <https://github.com/r-lib/keyring>

## Installation

Install the package from CRAN:

``` r
install.packages("oskeyring")
```

To install the development version from GitHub run:

``` r
pak::pak("r-lib/oskeyring")
```

## Usage

``` r
library(oskeyring)
```

Most oskeyring functions are not portable, and only work on one
operating system (OS). The functions that do not use the system
credential store can be used on all OSes. E.g. `macos_item()` and
`windows_item()` are portable. Calling a function on the wrong OS will
throw an `oskeyring_bad_os_error` error.

oskeyring follows the API of the OS closely, and it has a different set
of functions on Windows and macOS. E.g. the macOS API can search for
keychain items based on item attributes, but there is no similar API on
Windows, so oskeyring does not have a `windows_item_search()` function.

### Windows Credential Store

oskeyring uses the API defined in
[`wincred.h`](https://learn.microsoft.com/en-us/windows/win32/api/wincred/)
on Windows. The Windows credential store contains various credential
types. The ones supported by oskeyring are:

``` r
windows_item_types()
```

    #> [1] "generic"                 "domain_password"        
    #> [3] "domain_certificate"      "domain_visible_password"

`windows_item_write()` adds or updates a credential in the credential
store. It takes objects created with `windows_item()` :

``` r
it <- windows_item("secret", "my-host-password")
it
```

    #> <oskeyring_windows_item: generic>
    #>  target_name: my-host-password
    #>  persist: local_machine
    #>  credential_blob: <-- hidden -->

``` r
windows_item_write(it)
```

`windows_item_read()` reads a credential from the credential store, the
return value includes the secret as well:

``` r
windows_item_read("my-host-password")
```

``` r
#> <oskeyring_windows_item: generic>
#>  target_name: my-host-password
#>  persist: local_machine
#>  credential_blob: <-- hidden -->
```

`windows_item_enumerate()` lists all credentials that match a prefix.

`windows_item_delete()` deletes a credential.

See more in the manual: `?windows_credentials`.

### macOS Keychain Services

#### Keychain items

oskeyring uses the [Keychain
API](https://developer.apple.com/documentation/security/keychain_services)
on macOS. macOS keychains can store various classes of items. The item
classes supported by oskeyring are:

``` r
macos_item_classes()
```

    #> [1] "generic_password"  "internet_password"

`macos_item_add()` adds a new item to a keychain. It takes objects
created with `macos_item()`:

``` r
it <- macos_item(
  "secret",
  list(service = "My service", account = "Gabor"),
  class = "generic_password"
)
it
```

    #> <oskeyring_macos_item: generic_password>
    #>  account: Gabor
    #>  service: My service
    #>  value: <-- hidden -->

Items contain the secret itself, and a set of attributes, that depends
on the item class. See `?macos_keychain` for the list of attributes for
each class.

``` r
macos_item_add(it)
```

`macos_item_search()` searches for a keychain item:

``` r
macos_item_search(attributes = list(service = "My service"))
```

``` r
#> [[1]]
#> <oskeyring_macos_item: generic_password>
#>  account: Gabor
#>  creation_date: 2020-10-21 10:01:44
#>  label: My service
#>  modification_date: 2020-10-21 10:01:44
#>  service: My service
```

It does not return the secret itself, unless it is called with
`return_data = TRUE`. This possibly prompts the user for a password.

`macos_item_update()` updates the attributes of existing Keychain items.

`macos_item_delete()` deletes one or more Keychain items.

#### Keychains

macOS supports multiple keychains. There is always a default keychain,
and this is what oskeyring uses by default as well. There is also a
keychain search list, where secrets are looked up by default, and this
can contain multiple keychains.

`macos_item_*()` functions have a `keychain` argument to direct or
restrict the operation to a specific keychain.

`macos_keychain_create()` creates a new keychain.

`macos_keychain_list()` lists all keychains on the search list.

See more about macOS keychains in the manual: `?macos_keychain`.

## Code of Conduct

Please note that the oskeyring project is released with a [Contributor
Code of
Conduct](https://r-lib.github.io/oskeyring/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

## License

MIT © [RStudio](https://github.com/rstudio)

utils::globalVariables(c(
  "oskeyring_macos_add",
  "oskeyring_macos_delete",
  "oskeyring_macos_search",
  "oskeyring_macos_update",
  "oskeyring_macos_keychain_create",
  "oskeyring_macos_keychain_delete",
  "oskeyring_macos_keychain_is_locked",
  "oskeyring_macos_keychain_list",
  "oskeyring_macos_keychain_lock",
  "oskeyring_macos_keychain_unlock",
  "oskeyring_windows_write",
  "oskeyring_windows_read",
  "oskeyring_windows_delete",
  "oskeyring_windows_enumerate"
))

`%||%` <- function(l, r) if (is.null(l)) r else l

is_string <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x)
}

is_string_or_raw <- function(x) {
  is_string(x) || is.raw(x)
}

is_flag <- function(x) {
  is.logical(x) && length(x) == 1 && !is.na(x)
}

is_named_list <- function(x) {
  nms <- names(x)
  is.list(x) && length(nms) == length(x) && !any(is.na(nms)) && !any(nms == "")
}

# nocov start
ask_pass <- function(prompt = "Password: ") {
  res <- askpass::askpass(prompt)
  res <- res %||% askpass::askpass(prompt)
  enc2utf8(res)
}
# nocov end

os_check <- function(which = c("macOS", "Windows", "Linux")) {
  which <- match.arg(which)
  os <- get_os()
  if (os != tolower(which)) {
    err <- structure(
      list(
        message = paste0(
          "Unsupported OS. This function only works on ",
          which,
          "."
        )
      ),
      class = c("oskeyring_bad_os_error", "error", "condition")
    )
    stop(err)
  }
}

get_os <- function() {
  if (.Platform$OS.type == "windows") {
    "windows"
  } else if (Sys.info()["sysname"] == "Darwin") {
    "macos"
  } else if (Sys.info()[["sysname"]] == "Linux") {
    "linux"
  } else {
    "other"
  }
}

#' Query and manipulate the macOS Keychain
#'
#' @description
#' `macos_item_*` functions add, delete, update and search Keychain items.
#'
#' `macos_keychain_*` functions create, delete, list, lock, unlock
#' keychains.
#'
#' `macos_item_classes()` lists the supported Keychain item classes.
#' `macos_item_attr()` lists the supported attributes for these classes.
#' `macos_item_match_options()` lists the options supported by the
#' `match` argument of `macos_item_search()`.
#'
#' @details
#' # Keychain items
#'
#' `macos_item_classes()` returns the currently supported Keychain item
#' classes.
#'
#' ```{r}
#' macos_item_classes()
#' ```
#'
#' `macos_item()` creates a new Keychain item. See the next section about
#' the attributes that are supported for the various item types.
#'
#' ```{r}
#' it <- macos_item("secret", list(service = "My service", account = "Gabor"))
#' it
#' ```
#'
#' `macos_item_add()` adds an item to the keychain. If there is already an
#' item with the same primary keys, then it will error.
#'
#' ```{r, include = FALSE, eval = get_os() == "macos"}
#' # Remove, to make sure that 'add' works
#' tryCatch(
#'   macos_item_delete(attributes = list(service = "My service")),
#'   error = function(e) NULL
#' )
#' ```
#'
#' ```{r, eval = get_os() == "macos"}
#' macos_item_add(it)
#' ```
#'
#' `macos_item_search()` searches for Keychain items. If `return_data` is
#' `TRUE` then it also returns the secret data. Returning the secret data
#' might create a password entry dialog. If `return_data` is `TRUE` then
#' you need to set the `limit` match condition to a (small) finite number.
#'
#' ```{r, eval = get_os() == "macos"}
#' macos_item_search(attributes = list(service = "My service"))
#' ```
#'
#' `macos_item_update()` updates existing Keychain items.
#'
#' ```{r, eval = get_os() == "macos"}
#' macos_item_update(
#'   attributes = list(service = "My service", account = "Gabor"),
#'   update = list(account = "Gabor Csardi")
#' )
#' macos_item_search(attributes = list(service = "My service"))
#' ```
#'
#' `macos_item_delete()` deletes one or more Keychain items. Note that
#' all matching items will be deleted.
#'
#' ```{r, eval = get_os() == "macos"}
#' macos_item_delete(attributes = list(service = "My service"))
#' macos_item_search(attributes = list(service = "My service"))
#' ```
#'
#' ## Keychain Item Attributes
#'
#' * The set of supported attributes depends on the class of the item.
#' * oskeyring supports the following item classes currently:
#'   `r paste(macos_item_classes(), collapse = ", ")`.
#' * A subset of the attributes form a _primary key_. It is not possible
#'   to add more than one item with the same primary key. See the
#'   primary keys for the various classes below.
#' * oskeyring does not currently support all attributes that the
#'   Keychain Services AIP supports.
#' * Some attributes are read-only. If you try to set them when adding
#'   or updating items, they will be ignored.
#' * If an attribute is not included in the return value of
#'   `macos_item_search()` then it is not set, and its default value is in
#'   effect.
#'
#' ### Attributes for generic passwords
#'
#' `r item_list(macos_item_attr()[["generic_password"]])`
#'
#' ### Attributes for internet passwords
#'
#' `r item_list(macos_item_attr()[["internet_password"]])`
#'
#' # Search Parameters
#'
#' osxkeychain only supports a limited set of search parameters.
#' You can provide these for `macos_item_search()` as the `match` argument:
#'
#' `r item_list(macos_item_match_options())`
#'
#' # Keychains
#'
#' macOs supports multiple keychains.
#' There is always a default keychain, which is the user's login keychain,
#' unless configured differently.
#' There is also a keychain search list.
#' Keychains may belong into four non-exclusive categories, see the
#' `domain` argument of `macos_keychain_list()`.
#' A keychain is stored in an encrypted file on the disk, see the first
#' column of the output of `macos_keychain_list()`.
#'
#' `macos_item_*()` functions have a `keychain` argument to direct or
#' restrict the operation to a single keychain only. These are the defaults:
#' * `macos_item_add()` adds the item to the default keychain.
#' * `macos_item_search()` searches all keychains in the search list.
#' * `macos_item_update()` updates matching items on all keychains in the
#'   search list.
#' * `macos_item_delete()` deletes matching items from all keychains in the
#'   search list.
#'
#' `macos_keychain_create()` creates a new keychain.
#'
#' `macos_keychain_list()` lists all keychains on the search list.
#'
#' ```r
#' new <- "~/Library/Keychains/test.keychain-db"
#' macos_keychain_create(new, password = "secret")
#' macos_keychain_list()
#' ```
#'
#' ```r
#' ##                                                     path is_unlocked
#' ## 1 /Users/gaborcsardi/Library/Keychains/login.keychain-db        TRUE
#' ## 2 /Users/gaborcsardi/Library/Keychains/shiny.keychain-db       FALSE
#' ## 3  /Users/gaborcsardi/Library/Keychains/test.keychain-db        TRUE
#' ## 4                     /Library/Keychains/System.keychain       FALSE
#' ##   is_readable is_writeable
#' ## 1        TRUE         TRUE
#' ## 2        TRUE        FALSE
#' ## 3        TRUE         TRUE
#' ## 4        TRUE        FALSE
#' ```
#'
#' `macos_keychain_lock()` locks a keychain.
#' `macos_keychain_unlock()` unlocks a keychain.
#' `macos_keychain_is_locked()` checks if a keychain is locked.
#'
#' ```r
#' macos_keychain_lock(new)
#' macos_keychain_is_locked(new)
#' ```
#'
#' ```r
#' ## [1] TRUE
#' ```
#'
#' ```r
#' macos_keychain_unlock(new, password = "secret")
#' macos_keychain_is_locked(new)
#' ```
#'
#' ```r
#' ## [1] FALSE
#' ```
#'
#' `macos_keychain_delete()` deletes a keychain: it removes it from the
#' search list and deletes the data from the disk. It currently refuses to
#' delete the user's login keychain and the system keychain. Use Keychain
#' Access instead if you want to delete these. (Only do this if you are
#' aware of the bad consequences.)
#'
#' ```r
#' macos_keychain_delete(new)
#' macos_keychain_list()
#' ```
#'
#' ```r
#' ##                                                     path is_unlocked
#' ## 1 /Users/gaborcsardi/Library/Keychains/login.keychain-db        TRUE
#' ## 2 /Users/gaborcsardi/Library/Keychains/shiny.keychain-db       FALSE
#' ## 3                     /Library/Keychains/System.keychain       FALSE
#' ##   is_readable is_writeable
#' ## 1        TRUE         TRUE
#' ## 2        TRUE        FALSE
#' ## 3        TRUE        FALSE
#' ```
#'
#' @return `macos_item_classes()` returns a character vector, the
#' names of the supported keychain item classes.
#'
#' @seealso The Keychain Services API documentation at
#' <https://developer.apple.com/documentation/security/keychain_services>.
#' @export
#' @rdname macos_keychain
#' @name macos_keychain
#' @examples
#' # See above
macos_item_classes <- function() {
  # TODO: support the rest
  # c("generic_password", "internet_password", "certificate", "key", "identity")
  c("generic_password", "internet_password")
}

# ------------------------------------------------------------------------
# Keychain items
# ------------------------------------------------------------------------

#' @param class Item class, see [macos_item_classes()] for possible values.
#' @param value Value of the item, a password, key or certificate. It must
#' a raw vector or a string. If it is a string, then it is converted to
#' UTF-8.
#' @param attributes Item class dependent attributes in a named list.
#' See possible entries below
#' @return `macos_item()` returns a new `oskeyring_macos_item`
#' object.
#'
#' @export
#' @rdname macos_keychain

macos_item <- function(value, attributes = list(), class = "generic_password") {
  stopifnot(
    class %in% macos_item_classes(),
    is_string(value),
    is_macos_attributes(attributes, class)
  )

  if (is.character(value)) value <- charToRaw(enc2utf8(value))

  structure(
    list(class = class, value = value, attributes = attributes),
    class = "oskeyring_macos_item"
  )
}

format_attr <- function(x) {
  type <- vapply(x, typeof, character(1))
  x <- lapply(x, format)
  x <- ifelse(type != "raw", x, paste0("[raw] ", format(as.list(x))))
  nc <- nchar(x)
  ifelse(type != "raw" | nc <= 60, x, paste(substr(x, 1, 56), "..."))
}

#' @export
format.oskeyring_macos_item <- function(x, ...) {
  attr <- x$attributes[sort(names(x$attributes))]
  c(
    paste0("<oskeyring_macos_item: ", x$class, ">"),
    paste0(" ", names(attr), ": ", format_attr(attr)),
    if (!is.null(x$value)) " value: <-- hidden -->"
  )
}

#' @export

print.oskeyring_macos_item <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @param item Keychain item, creted via [macos_item()] or returned
#' by oskeyking itself.
#' @param keychain Select an alternative keychain, instead of the default.
#' Not implemented yet.
#' @return `macos_item_add()` returns `NULL`, invisibly.
#'
#' @export
#' @rdname macos_keychain

macos_item_add <- function(item, keychain = NULL) {
  os_check("macOS")
  stopifnot(
    inherits(item, "oskeyring_macos_item"),
    is_string(keychain) || is.null(keychain)
  )
  invisible(call_with_cleanup(oskeyring_macos_add, item, keychain))
}

#' @param class Type of items to search, see [macos_item_classes()] for
#' possible values.
#' @param attributes Narrow the search by indicating the attributes that
#' the found item or items should have.
#' @param match Condition the search in a variety of ways. For example, you
#' can limit the results to a specific number of items, control case
#' sensitivity when matching string attributes, etc. See 'Search parameters'
#' below.
#' @param return_data Whether to include the secret data in the
#' search result. If this is set to `TRUE`, then you'll have to set the
#' `limit` parameter (in the `match` argument) to a finite value.
#' If this is `TRUE`, then macOS will prompt you for passwords if necessary.
#' You might get multiple password prompts, if you set `limit` to a larger
#' than one value.
#' @return `macos_item_search()` returns a list of keychain items.
#'
#' @export
#' @rdname macos_keychain

macos_item_search <- function(
  class = "generic_password",
  attributes = list(),
  match = list(),
  return_data = FALSE,
  keychain = NULL
) {
  os_check("macOS")
  stopifnot(
    class %in% macos_item_classes(),
    is_macos_attributes(attributes, class),
    is_macos_match(match),
    is_flag(return_data),
    is_string(keychain) || is.null(keychain)
  )
  call_with_cleanup(
    oskeyring_macos_search,
    class,
    attributes,
    match,
    return_data,
    keychain
  )
}

#' @param update Named list specifying the new values of attributes.
#' @return `macos_item_update()` returns `NULL`, invisibly.
#'
#' @export
#' @rdname macos_keychain

macos_item_update <- function(
  class = "generic_password",
  attributes = list(),
  match = list(),
  update = list(),
  keychain = NULL
) {
  os_check("macOS")
  stopifnot(
    class %in% macos_item_classes(),
    is_macos_attributes(attributes, class),
    is_macos_match(match),
    is_macos_attributes(update, class),
    is_string(keychain) || is.null(keychain)
  )

  invisible(
    call_with_cleanup(
      oskeyring_macos_update,
      class,
      attributes,
      match,
      update,
      keychain
    )
  )
}

#' @return `macos_item_delete()` returns `NULL`, invisibly.
#' @export
#' @rdname macos_keychain

macos_item_delete <- function(
  class = "generic_password",
  attributes = list(),
  match = list(),
  keychain = NULL
) {
  os_check("macOS")
  stopifnot(
    class %in% macos_item_classes(),
    is_macos_attributes(attributes, class),
    is_macos_match(match),
    is_string(keychain) || is.null(keychain)
  )

  invisible(
    call_with_cleanup(
      oskeyring_macos_delete,
      class,
      attributes,
      match,
      keychain
    )
  )
}

# ------------------------------------------------------------------------
# Access control
# ------------------------------------------------------------------------

# TODO

# ------------------------------------------------------------------------
# Keychains
# ------------------------------------------------------------------------

#' @export
#' @param keychain Keychain to use. `NULL` means the default one.
#' @param password Password to unlock the keychain, or new password to
#' set when creating a new keychain. May be `NULL` in interactive
#' sessions, to force a secure password dialog.
#' @return `macos_keychain_create()` returns `NULL`, invisibly.
#'
#' @rdname macos_keychain

macos_keychain_create <- function(keychain, password = NULL) {
  os_check("macOS")
  stopifnot(
    is_string(keychain),
    is_string(password) || is.null(password)
  )

  password <- password %||% ask_pass("Keychain password: ")
  file <- macos_keychain_file(keychain)
  invisible(call_with_cleanup(oskeyring_macos_keychain_create, file, password))
}

#' @param domain The preference domain from which you wish to retrieve
#' the keychain search list:
#' * `"all"`: include all keychains currently on the search list,
#' * `"user"`: user preference domain,
#' * `"system"`: system or daemon preference domain,
#' * `"common"`: keychains common to everyone,
#' * `"dynamic"`: dynamic search list (typically provided by removable
#' keychains such as smart cards).
#' @return `macos_keychain_list()` returns a data frame with columns:
#' * `path`: Path to the file of the keychain.
#' * `is_locked`: Whether the keychain is locked.
#' * `is_readable`: Whether the keychain is readable by the user.
#' * `is_writeable`: Whether the keychain is writeable by the user.
#'
#' @export
#' @rdname macos_keychain

macos_keychain_list <- function(
  domain = c(
    "all",
    "user",
    "system",
    "common",
    "dynamic"
  )
) {
  os_check("macOS")
  domain <- match.arg(domain)
  ret <- call_with_cleanup(oskeyring_macos_keychain_list, domain)
  data.frame(
    stringsAsFactors = FALSE,
    path = ret[[1]],
    is_unlocked = ret[[2]],
    is_readable = ret[[3]],
    is_writeable = ret[[4]]
  )
}

#' @return `macos_keychain_delete()` returns `NULL`, invisibly.
#' @export
#' @rdname macos_keychain

macos_keychain_delete <- function(keychain) {
  os_check("macOS")
  stopifnot(
    is_string(keychain)
  )
  if (grepl("^login\\.keychain", basename(keychain))) {
    stop("Refusing to delete the login keychain")
  }
  if (grepl("^System\\.keychain", basename(keychain))) {
    stop("Refusing to delete the system keychain")
  }
  file <- macos_keychain_file(keychain)
  invisible(call_with_cleanup(oskeyring_macos_keychain_delete, file))
}

#' @return `macos_keychain_lock()` returns `NULL`, invisibly.
#' @export
#' @rdname macos_keychain

macos_keychain_lock <- function(keychain = NULL) {
  os_check("macOS")
  stopifnot(
    is_string(keychain) || is.null(keychain)
  )
  file <- macos_keychain_file(keychain)
  invisible(call_with_cleanup(oskeyring_macos_keychain_lock, file))
}

#' @return `macos_keychain_unlock()` returns `NULL`, invisibly.
#' @export
#' @rdname macos_keychain

macos_keychain_unlock <- function(keychain = NULL, password = NULL) {
  os_check("macOS")
  stopifnot(
    is_string(keychain) || is.null(keychain),
    is_string(password) || is.null(password)
  )

  password <- password %||% ask_pass("Password to unlock keychain: ")
  file <- macos_keychain_file(keychain)
  invisible(call_with_cleanup(oskeyring_macos_keychain_unlock, file, password))
}

#' @return `macos_keychain_is_locked()` returns `TRUE` or `FALSE`.
#' @export
#' @rdname macos_keychain

macos_keychain_is_locked <- function(keychain = NULL) {
  os_check("macOS")
  stopifnot(
    is_string(keychain) || is.null(keychain)
  )
  file <- macos_keychain_file(keychain)
  call_with_cleanup(oskeyring_macos_keychain_is_locked, file)
}

#' @return `macos_item_attr()` returns a list of lists of character
#' scalars, the description of keychain item attributes, for each
#' keychain item class.
#' @export
#' @rdname macos_keychain

macos_item_attr <- function() {
  list(
    generic_password = list(
      #      access,
      #      access_control,
      #      access_group,
      #      accessible,
      creation_date = paste0(
        "[.POSIXct(1)][read-only] The date the item was created."
      ),
      modification_date = paste0(
        "[.POSIXct(1)][read-only] The last time the item was updated."
      ),
      description = paste0(
        "[character(1)] User-visible string describing this kind of",
        "item (for example, 'Disk image password')."
      ),
      comment = "[character(1)] User-editable comment for this item.",
      #      "creator",
      #      "type",
      label = "[character(1)] User-visible label for this item.",
      is_invisible = paste0(
        "[logical(1)] `TRUE` if the item is invisible (that is, should ",
        "not be displayed)."
      ),
      is_negative = paste0(
        "[logical(1)] Indicates whether there is a valid password ",
        "associated with this keychain item. This is useful if your ",
        "application doesn't want a password for some particular service ",
        "to be stored in the keychain, but prefers that it always be ",
        "entered by the user."
      ),
      account = "[character(1)][key] Account name.",
      service = paste0(
        "[character(1)][key] The service associated with this item."
      ),
      generic = "[character(1)] User-defined attribute.",
      synchronizable = paste0(
        "[logical(1)] Indicates whether the item in question is ",
        "synchronized to other devices through iCloud."
      )
    ),
    internet_password = list(
      #      access,
      #      access_group,
      #      accessible,
      creation_date = paste0(
        "[.POSIXct(1)][read-only] The date the item was created."
      ),
      modification_date = paste0(
        "[.POSIXct(1)][read-only] The last time the item was updated."
      ),
      description = paste0(
        "[character(1)] User-visible string describing this kind of",
        "item (for example, 'Disk image password')."
      ),
      comment = "[character(1)] User-editable comment for this item.",
      #      "creator",
      #      "type",
      label = "[character(1)] User-visible label for this item.",
      is_invisible = paste0(
        "[logical(1)] `TRUE` if the item is invisible (that is, should ",
        "not be displayed)."
      ),
      is_negative = paste0(
        "[logical(1)] Indicates whether there is a valid password ",
        "associated with this keychain item. This is useful if your ",
        "application doesn't want a password for some particular service ",
        "to be stored in the keychain, but prefers that it always be ",
        "entered by the user."
      ),
      account = "[character(1)][key] Account name.",
      synchronizable = paste0(
        "[logical(1)] Indicates whether the item in question is ",
        "synchronized to other devices through iCloud."
      ),
      security_domain = "[character(1)][key] The item's security domain.",
      server = paste0(
        "[character(1)][key] Contains the server's domain name or IP ",
        "address."
      ),
      protocol = "[character(1)][key] The protocol for this item.",
      authentication_type = "character[1][key] Authentication type.",
      port = "[integer(1)][key] Internet port number.",
      path = paste0(
        "[character(1)][key] A path, typically the path component of ",
        "the URL"
      )
    )
  )
}

#' @return `macos_item_match_options()` returns a list of character
#' scalars, the description of the supported match options.
#' @export
#' @rdname macos_keychain

macos_item_match_options <- function() {
  list(
    ## TODO: authentication_context,
    ## TODO: how does use_authentication_ui work?
    ## TODO: use_operation_prompt
    ## policy,
    ## issuers,
    ## email_address_if_present,
    ## subject_contains,
    ## subject_starts_with,
    ## subject_ends_with,
    ## subject_whole_string,
    ## TODO: case_insensitive does not work? Only for certs?
    ## TODO: diacritic_insensitive similar
    ## TODO: width_insensitive similar
    ## trusted_only,
    ## valid_on_date,
    limit = paste0(
      "[numeric(1)] This value specifies the maximum number of results ",
      "to return or otherwise act upon. Use `Inf` to specify all ",
      "matching items."
    )
  )
}

# ------------------------------------------------------------------------
# Internals
# ------------------------------------------------------------------------

macos_keychain_file <- function(keychain = NULL) {
  if (!is.null(keychain)) {
    enc2utf8(normalizePath(keychain, mustWork = FALSE))
  }
}

item_list <- function(x) {
  nms <- names(x)
  x <- gsub("[", "\\[", x, fixed = TRUE)
  x <- gsub("]", "\\]", x, fixed = TRUE)
  paste0("* `", nms, "`: ", x, collapse = "\n")
}

is_macos_attributes <- function(attr, class) {
  if (!is_named_list(attr)) {
    return(FALSE)
  }
  good <- macos_item_attr()[[class]]
  bad <- setdiff(names(attr), names(good))
  if (length(bad)) {
    stop(
      "Unknown attributes for `",
      class,
      "`:",
      paste0("`", bad, "`", collapse = ", ")
    )
  }
  TRUE
}

is_macos_match <- function(x) {
  if (!is_named_list(x)) {
    return(FALSE)
  }
  mtch <- macos_item_match_options()
  bad <- setdiff(names(x), names(mtch))
  if (length(bad)) {
    stop(
      "Unknown match parameters :",
      paste0("`", bad, "`", collapse = ", ")
    )
  }
  TRUE
}

is_ci <- function() {
  Sys.getenv("CI", "") != ""
}

random_name <- function() {
  paste0("oskeyring-test-", basename(tempfile()))
}

osk_test_that <- function(os, desc, code) {
  if (is_ci() && tolower(os) == get_os()) {
    testthat::test_that(desc, {
      code
    })
  }
}

test_cleanup_macos <- function() {
  if (is_ci() && tolower(get_os()) == "macos") {
    empty_error <- function(x) grepl("not be found", x$message)
    tryCatch(
      macos_item_delete(
        "generic_password",
        attributes = list(comment = "oskeyring-test")
      ),
      error = function(err) if (!empty_error(err)) print(err)
    )
    tryCatch(
      macos_item_delete(
        "internet_password",
        attributes = list(comment = "oskeyring-test")
      ),
      error = function(err) if (!empty_error(err)) print(err)
    )
    tryCatch(
      {
        lst <- macos_keychain_list()
        tst <- grep("oskeyking-test", lst$path, value = TRUE)
        for (fn in tst) try(macos_keychain_delete(fn))
      },
      error = function(err) print(err)
    )
  }
}

test_cleanup_windows <- function() {
  if (is_ci() && tolower(get_os()) == "windows") {
    tryCatch(
      {
        its <- windows_item_enumerate("oskeyring-test-*")
        for (it in its) {
          try(windows_item_delete(it$target_name, it$type))
        }
      },
      error = function(err) print(err)
    )
  }
}

if (packageVersion("testthat") <= "2.5.0") {
  testthat::teardown(test_cleanup_macos())
} else {
  withr::defer(test_cleanup_macos(), teardown_env())
}

test_that("macos_item_classes", {
  expect_true(is.character(macos_item_classes()))
})

test_that("macos_item", {
  it <- macos_item("qwerty123", list(service = "service"))
  expect_s3_class(it, "oskeyring_macos_item")
})

test_that("format.oskeyring_macos_item", {
  it <- macos_item("qwerty123", list(service = "service"))
  fmt <- format(it)
  expect_false(any(grepl("qwerty123", fmt)))
})

test_that("print.oskeyring_macos_item", {
  it <- macos_item("qwerty123", list(service = "service"))
  out <- capture.output(print(it))
  expect_false(any(grepl("qwerty123", out)))
})

osk_test_that("macos", "macos_item_add", {
  service <- random_name()
  it <- macos_item(
    "secret!",
    list(service = service, description = "foo", comment = "oskeyring-test")
  )
  expect_null(macos_item_add(it))

  eit <- macos_item_search("generic_password", list(service = service))
  expect_equal(length(eit), 1)
  expect_equal(eit[[1]]$attributes$description, "foo")

  expect_null(
    macos_item_delete("generic_password", list(service = service))
  )
  expect_equal(macos_item_search(, list(service = service)), list())

  it2 <- macos_item(
    "secret2!",
    list(server = service, description = "bar", comment = "oskeyring-test"),
    "internet_password"
  )
  expect_null(macos_item_add(it2))

  eit2 <- macos_item_search("internet_password", list(server = service))
  expect_equal(length(eit2), 1)
  expect_equal(eit2[[1]]$attributes$description, "bar")

  expect_null(
    macos_item_delete("internet_password", list(server = service))
  )
  expect_equal(
    macos_item_search("internet_password", list(server = service)),
    list()
  )
})

osk_test_that("macos", "unsupported os", {
  it <- windows_item("foo", "bar")
  expect_error(
    windows_item_write(it),
    class = "oskeyring_bad_os_error"
  )
})

osk_test_that("macos", "macos_item_update", {
  service <- random_name()
  it <- macos_item(
    "secret!",
    list(service = service, description = "foo", comment = "oskeyring-test")
  )
  expect_null(macos_item_add(it))

  service2 <- paste0(service, "-2")
  macos_item_update(
    "generic_password",
    attributes = list(service = service),
    update = list(service = service2, description = "bar")
  )

  eit <- macos_item_search("generic_password", list(service = service2))
  expect_equal(length(eit), 1)
  expect_equal(eit[[1]]$attributes$description, "bar")
})

osk_test_that("macos", "keychains", {
  new <- paste0(
    "~/Library/Keychains/oskeyking-test-",
    basename(tempfile()),
    ".keychain-db"
  )
  expect_null(macos_keychain_create(new, password = "secret"))
  lst <- macos_keychain_list()
  expect_true(any(grepl(basename(new), lst$path, fixed = TRUE)))
  expect_null(macos_keychain_lock(new))
  expect_true(macos_keychain_is_locked(new))
  expect_null(macos_keychain_unlock(new, password = "secret"))
  expect_false(macos_keychain_is_locked(new))
})

osk_test_that("macos", "will not delete login/system keychain", {
  expect_error(
    macos_keychain_delete("foo/bar/login.keychain"),
    "Refusing to delete"
  )
  expect_error(
    macos_keychain_delete("foo/bar/System.keychain"),
    "Refusing to delete"
  )
})

test_that("item_list", {
  expect_equal(
    item_list(c(foo = "bar", foobar = "baz")),
    "* `foo`: bar\n* `foobar`: baz"
  )
})

test_that("is_macos_attributes", {
  expect_false(is_macos_attributes(list("a", "b")))
  expect_error(
    is_macos_attributes(list(foo = "bar"), "generic_password"),
    "Unknown attributes"
  )
})

test_that("is_macos_match", {
  expect_false(is_macos_match(list("a", "b")))
  expect_error(
    is_macos_match(list(foo = "bar")),
    "Unknown match parameters"
  )
})

#' Query and manipulate the Windows Credential Store
#'
#' @description
#' `windows_item_*` functions read, write, delete and list
#' credentials.
#'
#' @details
#' ## `windows_item_types()`
#'
#' `windows_item_types()` lists the currently supported credential
#' types.
#'
#' ```{r}
#' windows_item_types()
#' ```
#'
#' ## `windows_item()`
#'
#' `windows_item()` creates a Windows credential, that can be
#' then added to the credential store.
#'
#' ```{r}
#' it <- windows_item("secret", "my-host-password")
#' it
#' ```
#'
#' ## `windows_item_write()`
#'
#' Writes an item to the credential store.
#'
#' ```{r, include = FALSE, eval = get_os() == "windows"}
#' # Remove to make sure it is not there
#' tryCatch(
#'   windows_item_delete("my-host-password"),
#'   error = function(err) NULL
#' )
#' ```
#'
#' ```{r, eval = get_os() == "windows"}
#' windows_item_write(it)
#' ```
#'
#' ## `windows_item_read()`
#'
#' Reads a credential with the specified type and `target_name`.
#'
#' ```{r, eval = get_os() == "windows"}
#' windows_item_read("my-host-password")
#' ```
#'
#' ## `windows_item_enumerate()`
#'
#' List all credentials that match a prefix.
#'
#' ```{r, eval = get_os() == "windows"}
#' windows_item_enumerate(filter = "my-*")
#' ```
#'
#' ## `windows_item_delete()`
#'
#' Delete a credential:
#'
#' ```{r, eval = get_os() == "windows"}
#' windows_item_delete("my-host-password")
#' windows_item_enumerate(filter = "my-*")
#' ```
#'
#' @seealso The API documentation at
#' <https://learn.microsoft.com/en-us/windows/win32/api/wincred/>
#' @export
#' @return `windows_item_types()` returns a character vector, the
#' currently supported credential types.
#' @rdname windows_credentials
#' @name windows_credentials
#' @examples
#' # See above
windows_item_types <- function() {
  c(
    "generic",
    "domain_password",
    "domain_certificate",
    "domain_visible_password"
  )
}

#' @param credential_blob The secret credential, a password,
#' certificate or key. See also
#' <https://learn.microsoft.com/en-us/windows/win32/api/wincred/>
#' This can be a raw vector, or a string. If it is a string, then it
#' will be converted to Unicode, without the terminating zero.
#' It can also be `NULL`, to be used with the `preserve = TRUE`
#' argument of `windows_item_write()`.
#' @param target_name The name of the credential. The `target_name`
#' and `type` members uniquely identify the credential. This member
#' cannot be changed after the credential is created. Instead, the
#' credential with the old name should be deleted and the credential
#' with the new name created. This member cannot be longer than
#' `CRED_MAX_GENERIC_TARGET_NAME_LENGTH` (32767) characters.
#' This member is case-insensitive.
#' @param type The type of the credential. This member cannot be
#' changed after the credential is created. See `windows_item_types()`
#' for possible values.
#' @param comment If not `NULL`, then a string comment from the user
#' that describes this credential. This member cannot be longer than
#' `CRED_MAX_STRING_LENGTH` (256) characters. It is stored as a Unicode
#' string.
#' @param persist Defines the persistence of this credential.
#' * `"local_machine"`: The credential persists for all subsequent
#'   logon sessions on this same computer. It is visible to other
#'   logon sessions of this same user on this same computer and not
#'   visible to logon sessions for this user on other computers.
#' * `"session"`: The credential persists for the life of the logon
#'   session. It will not be visible to other logon sessions of this
#'   same user. It will not exist after this user logs off and back on.
#' * `"enterprise"`: The credential persists for all subsequent logon
#'   sessions on this same computer. It is visible to other logon
#'   sessions of this same user on this same computer and to logon
#'   sessions for this user on other computers.
#' @param attributes Application-defined attributes that are
#'   associated with the credential. This is `NULL` or a named list
#'   of raw or string vectors. String vectors are converted to
#'   Unicode, without the terminating zero. A credential can have at
#'   most 64 attributes, the names of the attributes cannot be
#'   longer than `CRED_MAX_STRING_LENGTH` (256) characters each, and
#'   the attributes themselves cannot be longer than
#'   `CRED_MAX_VALUE_SIZE` (256) bytes.
#' @param target_alias Alias for the `target_name` member.
#' This member can be read and written. It cannot be longer than
#' `CRED_MAX_STRING_LENGTH` (256) characters. It is stored in Unicode.
#' @param username `NULL` or the user name of the account used to
#' connect to `target_name`.
#' @return `windows_item()` returns an `oskeyring_windows_item`
#' object.
#' @export
#' @rdname windows_credentials

windows_item <- function(
  credential_blob,
  target_name,
  type = "generic",
  comment = NULL,
  persist = c(
    "local_machine",
    "session",
    "enterprise"
  ),
  attributes = list(),
  target_alias = NULL,
  username = NULL
) {
  persist <- match.arg(persist)
  stopifnot(
    is.null(credential_blob) || is_string_or_raw(credential_blob),
    is_string(target_name),
    is_string(type) && type %in% windows_item_types(),
    is.null(comment) || is_string(comment),
    is_named_list(attributes) &&
      all(vapply(attributes, function(x) is_string_or_raw(x), logical(1))),
    is.null(target_alias) || is_string(target_alias),
    is.null(username) || is_string(username)
  )

  structure(
    list(
      type = type,
      target_name = target_name,
      credential_blob = credential_blob,
      comment = comment,
      persist = persist,
      attributes = attributes,
      target_alias = target_alias,
      username = username
    ),
    class = "oskeyring_windows_item"
  )
}

#' @export

format.oskeyring_windows_item <- function(x, ...) {
  c(
    paste0("<oskeyring_windows_item: ", x$type, ">"),
    paste0(" target_name: ", x$target_name),
    if (!is.null(x$comment)) paste0(" comment: ", x$comment),
    paste0(" persist: ", x$persist),
    if (!is.null(x$target_alias)) paste0(" target_alias: ", x$target_alias),
    if (!is.null(x$username)) paste0(" username: ", x$username),
    if (length(x$attributes) > 0) {
      c(
        " attributes:",
        paste0("  ", names(x$attributes), ": ", format_attr(x$attributes))
      )
    },
    if (!is.null(x$credential_blob)) " credential_blob: <-- hidden -->"
  )
}

#' @export

print.oskeyring_windows_item <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @return `windows_item_read()` returns an `oskeyring_windows_item`
#' object.
#' @export
#' @rdname windows_credentials

windows_item_read <- function(target_name, type = "generic") {
  os_check("Windows")
  stopifnot(
    is_string(target_name),
    is_string(type) && type %in% windows_item_types()
  )

  target_name <- to_utf16(target_name)
  item <- call_with_cleanup(oskeyring_windows_read, target_name, type)

  windows_item_from_utf16(item)
}

#' @param item `oskeyring_windows_item` object to write.
#' @param preserve The credential BLOB from an existing credential
#' is preserved with the same credential name and credential type.
#' The `credential_blob` of the passed `oskeyring_windows_item`
#' object must be `NULL`.
#' @return `windows_item_write()` returns `NULL`, invisibly.
#' @export
#' @rdname windows_credentials

windows_item_write <- function(item, preserve = FALSE) {
  os_check("Windows")
  stopifnot(
    inherits(item, "oskeyring_windows_item"),
    is_flag(preserve)
  )

  encode <- function(x) {
    if (is.character(x)) {
      iconv(x, to = "UTF-16LE", toRaw = TRUE)[[1]]
    } else if (is.null(x) || is.raw(x)) {
      x
    } else {
      stop("Unsupported data type in Windows keychain item")
    }
  }
  item["credential_blob"] <- list(encode(item$credential_blob))
  item$target_name <- to_utf16(item$target_name)
  item["comment"] <- list(to_utf16(item$comment))
  item["attribute_names"] <- list(lapply(names(item$attributes), to_utf16))
  item["attributes"] <- list(lapply(item$attributes, encode))
  item["target_alias"] <- list(to_utf16(item$target_alias))
  item["username"] <- list(to_utf16(item$username))

  invisible(call_with_cleanup(oskeyring_windows_write, item, preserve))
}

#' @return `windows_item_delete()` returns `NULL`, invisibly.
#' @export
#' @rdname windows_credentials

windows_item_delete <- function(target_name, type = "generic") {
  os_check("Windows")
  stopifnot(
    is_string(target_name),
    is_string(type) && type %in% windows_item_types()
  )

  target_name <- to_utf16(target_name)
  invisible(call_with_cleanup(oskeyring_windows_delete, target_name, type))
}

#' @param filter If not `NULL`, then a string to filter the
#' credentials. Only credentials with a `target_name` matching the
#' filter will be returned. The filter specifies a name prefix
#' followed by an asterisk. For instance, the filter `"FRED*"` will
#' return all credentials with a `target_name` beginning with the
#' string `"FRED"`.
#' @param all Whether to use the `CRED_ENUMERATE_ALL_CREDENTIALS`
#' flag to enumerate all credentials. If this is `TRUE`, then `filter`
#' must be `NULL`. If this is `TRUE`, then the target name of each
#' credential is returned in the `"namespace:attribute=target`" format.
#' @return `windows_item_enumerate()` returns a list of
#' `oskeyring_windows_item` items.
#' @export
#' @rdname windows_credentials

windows_item_enumerate <- function(filter = NULL, all = FALSE) {
  os_check("Windows")
  stopifnot(
    is.null(filter) || is_string(filter),
    is_flag(all)
  )

  if (!is.null(filter)) filter <- to_utf16(filter)
  items <- call_with_cleanup(oskeyring_windows_enumerate, filter, all)
  lapply(items, windows_item_from_utf16)
}

windows_item_from_utf16 <- function(item) {
  item$target_name <- from_utf16(list(item$target_name))
  item["comment"] <- list(from_utf16(list(item$comment)))
  item["target_alias"] <- list(from_utf16(list(item$target_alias)))
  item["username"] <- list(from_utf16(list(item$username)))

  if (length(item$attributes)) {
    attr <- item$attributes
    names(attr) <- from_utf16(item$attribute_names)
    item["attributes"] <- list(attr)
  } else {
    item["attributes"] <- list(NULL)
  }
  item$attribute_names <- NULL

  item
}

to_utf16 <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }
  stopifnot(is_string(x))
  c(
    iconv(enc2utf8(x), from = "UTF-8", to = "UTF-16LE", toRaw = TRUE)[[1]],
    raw(2)
  )
}

from_utf16 <- function(x) {
  if (length(x) == 1 && is.null(x[[1]])) {
    return(NULL)
  }
  iconv(x, from = "UTF-16LE", to = "UTF-8")
}

if (packageVersion("testthat") <= "2.5.0") {
  testthat::teardown(test_cleanup_windows())
} else {
  withr::defer(test_cleanup_windows(), teardown_env())
}

test_that("windows_item_types", {
  expect_true(is.character(windows_item_types()))
})

test_that("windows_item", {
  it <- windows_item("secret", "target")
  expect_s3_class(it, "oskeyring_windows_item")
})

test_that("format.windows_item", {
  attr <- list(key1 = "value1", key2 = "value2")
  it <- windows_item("qwerty123", "target", attributes = attr)
  fmt <- format(it)
  expect_false(any(grepl("qwerty123", fmt)))
  expect_true(any(grepl("key1.*value1", fmt)))
  expect_true(any(grepl("key2.*value2", fmt)))
})

test_that("print.windows_item", {
  it <- windows_item("qwerty123", "target")
  out <- capture.output(print(it))
  expect_false(any(grepl("qwerty123", out)))
})

osk_test_that("windows", "windows_item_write", {
  target <- random_name()
  it <- windows_item(
    "secret!",
    target,
    attributes = list(key1 = charToRaw("value1"))
  )
  expect_null(windows_item_write(it))

  eit <- windows_item_read(target)
  expect_s3_class(eit, "oskeyring_windows_item")
  expect_equal(
    iconv(list(eit$credential_blob), "UTF-16LE", "UTF-8"),
    "secret!"
  )

  it2 <- windows_item(NULL, target)
  expect_null(windows_item_write(it2, preserve = TRUE))
  eit2 <- windows_item_read(target)

  lst <- windows_item_enumerate("oskeyring-test-*")
  tgs <- vapply(lst, "[[", character(1), "target_name")
  expect_true(target %in% tgs)

  expect_null(windows_item_delete(target))

  expect_error(
    windows_item_read(target),
    "Windows credential store error"
  )
})

osk_test_that("windows", "windows_item_write #2", {
  it <- windows_item("x", random_name())
  it$attributes <- list(x = 1L)
  expect_error(
    windows_item_write(it)
  )
})

osk_test_that("windows", "unsupported os", {
  it <- macos_item("foobar")
  expect_error(
    macos_item_add(it),
    class = "oskeyring_bad_os_error"
  )
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/windows.R
\name{windows_credentials}
\alias{windows_credentials}
\alias{windows_item_types}
\alias{windows_item}
\alias{windows_item_read}
\alias{windows_item_write}
\alias{windows_item_delete}
\alias{windows_item_enumerate}
\title{Query and manipulate the Windows Credential Store}
\usage{
windows_item_types()

windows_item(
  credential_blob,
  target_name,
  type = "generic",
  comment = NULL,
  persist = c("local_machine", "session", "enterprise"),
  attributes = list(),
  target_alias = NULL,
  username = NULL
)

windows_item_read(target_name, type = "generic")

windows_item_write(item, preserve = FALSE)

windows_item_delete(target_name, type = "generic")

windows_item_enumerate(filter = NULL, all = FALSE)
}
\arguments{
\item{credential_blob}{The secret credential, a password,
certificate or key. See also
\url{https://learn.microsoft.com/en-us/windows/win32/api/wincred/}
This can be a raw vector, or a string. If it is a string, then it
will be converted to Unicode, without the terminating zero.
It can also be \code{NULL}, to be used with the \code{preserve = TRUE}
argument of \code{windows_item_write()}.}

\item{target_name}{The name of the credential. The \code{target_name}
and \code{type} members uniquely identify the credential. This member
cannot be changed after the credential is created. Instead, the
credential with the old name should be deleted and the credential
with the new name created. This member cannot be longer than
\code{CRED_MAX_GENERIC_TARGET_NAME_LENGTH} (32767) characters.
This member is case-insensitive.}

\item{type}{The type of the credential. This member cannot be
changed after the credential is created. See \code{windows_item_types()}
for possible values.}

\item{comment}{If not \code{NULL}, then a string comment from the user
that describes this credential. This member cannot be longer than
\code{CRED_MAX_STRING_LENGTH} (256) characters. It is stored as a Unicode
string.}

\item{persist}{Defines the persistence of this credential.
\itemize{
\item \code{"local_machine"}: The credential persists for all subsequent
logon sessions on this same computer. It is visible to other
logon sessions of this same user on this same computer and not
visible to logon sessions for this user on other computers.
\item \code{"session"}: The credential persists for the life of the logon
session. It will not be visible to other logon sessions of this
same user. It will not exist after this user logs off and back on.
\item \code{"enterprise"}: The credential persists for all subsequent logon
sessions on this same computer. It is visible to other logon
sessions of this same user on this same computer and to logon
sessions for this user on other computers.
}}

\item{attributes}{Application-defined attributes that are
associated with the credential. This is \code{NULL} or a named list
of raw or string vectors. String vectors are converted to
Unicode, without the terminating zero. A credential can have at
most 64 attributes, the names of the attributes cannot be
longer than \code{CRED_MAX_STRING_LENGTH} (256) characters each, and
the attributes themselves cannot be longer than
\code{CRED_MAX_VALUE_SIZE} (256) bytes.}

\item{target_alias}{Alias for the \code{target_name} member.
This member can be read and written. It cannot be longer than
\code{CRED_MAX_STRING_LENGTH} (256) characters. It is stored in Unicode.}

\item{username}{\code{NULL} or the user name of the account used to
connect to \code{target_name}.}

\item{item}{\code{oskeyring_windows_item} object to write.}

\item{preserve}{The credential BLOB from an existing credential
is preserved with the same credential name and credential type.
The \code{credential_blob} of the passed \code{oskeyring_windows_item}
object must be \code{NULL}.}

\item{filter}{If not \code{NULL}, then a string to filter the
credentials. Only credentials with a \code{target_name} matching the
filter will be returned. The filter specifies a name prefix
followed by an asterisk. For instance, the filter \code{"FRED*"} will
return all credentials with a \code{target_name} beginning with the
string \code{"FRED"}.}

\item{all}{Whether to use the \code{CRED_ENUMERATE_ALL_CREDENTIALS}
flag to enumerate all credentials. If this is \code{TRUE}, then \code{filter}
must be \code{NULL}. If this is \code{TRUE}, then the target name of each
credential is returned in the \verb{"namespace:attribute=target}" format.}
}
\value{
\code{windows_item_types()} returns a character vector, the
currently supported credential types.

\code{windows_item()} returns an \code{oskeyring_windows_item}
object.

\code{windows_item_read()} returns an \code{oskeyring_windows_item}
object.

\code{windows_item_write()} returns \code{NULL}, invisibly.

\code{windows_item_delete()} returns \code{NULL}, invisibly.

\code{windows_item_enumerate()} returns a list of
\code{oskeyring_windows_item} items.
}
\description{
\verb{windows_item_*} functions read, write, delete and list
credentials.
}
\details{
\subsection{\code{windows_item_types()}}{

\code{windows_item_types()} lists the currently supported credential
types.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{windows_item_types()
#> [1] "generic"                 "domain_password"        
#> [3] "domain_certificate"      "domain_visible_password"
}\if{html}{\out{</div>}}
}

\subsection{\code{windows_item()}}{

\code{windows_item()} creates a Windows credential, that can be
then added to the credential store.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{it <- windows_item("secret", "my-host-password")
it
#> <oskeyring_windows_item: generic>
#>  target_name: my-host-password
#>  persist: local_machine
#>  credential_blob: <-- hidden -->
}\if{html}{\out{</div>}}
}

\subsection{\code{windows_item_write()}}{

Writes an item to the credential store.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{windows_item_write(it)
}\if{html}{\out{</div>}}
}

\subsection{\code{windows_item_read()}}{

Reads a credential with the specified type and \code{target_name}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{windows_item_read("my-host-password")
}\if{html}{\out{</div>}}
}

\subsection{\code{windows_item_enumerate()}}{

List all credentials that match a prefix.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{windows_item_enumerate(filter = "my-*")
}\if{html}{\out{</div>}}
}

\subsection{\code{windows_item_delete()}}{

Delete a credential:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{windows_item_delete("my-host-password")
windows_item_enumerate(filter = "my-*")
}\if{html}{\out{</div>}}
}
}
\examples{
# See above
}
\seealso{
The API documentation at
\url{https://learn.microsoft.com/en-us/windows/win32/api/wincred/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oskeyring-package.R
\docType{package}
\name{oskeyring-package}
\alias{oskeyring}
\alias{oskeyring-package}
\title{oskeyring: Raw System Credential Store Access from R}
\description{
Aims to support all features of the system credential store, including non-portable ones. Supports 'Keychain' on 'macOS', and 'Credential Manager' on 'Windows'. See the 'keyring' package if you need a portable 'API'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lib/oskeyring#readme}
  \item \url{https://r-lib.github.io/oskeyring/}
  \item Report bugs at \url{https://github.com/r-lib/oskeyring/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{csardi.gabor@gmail.com}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/macos.R
\name{macos_keychain}
\alias{macos_keychain}
\alias{macos_item_classes}
\alias{macos_item}
\alias{macos_item_add}
\alias{macos_item_search}
\alias{macos_item_update}
\alias{macos_item_delete}
\alias{macos_keychain_create}
\alias{macos_keychain_list}
\alias{macos_keychain_delete}
\alias{macos_keychain_lock}
\alias{macos_keychain_unlock}
\alias{macos_keychain_is_locked}
\alias{macos_item_attr}
\alias{macos_item_match_options}
\title{Query and manipulate the macOS Keychain}
\usage{
macos_item_classes()

macos_item(value, attributes = list(), class = "generic_password")

macos_item_add(item, keychain = NULL)

macos_item_search(
  class = "generic_password",
  attributes = list(),
  match = list(),
  return_data = FALSE,
  keychain = NULL
)

macos_item_update(
  class = "generic_password",
  attributes = list(),
  match = list(),
  update = list(),
  keychain = NULL
)

macos_item_delete(
  class = "generic_password",
  attributes = list(),
  match = list(),
  keychain = NULL
)

macos_keychain_create(keychain, password = NULL)

macos_keychain_list(domain = c("all", "user", "system", "common", "dynamic"))

macos_keychain_delete(keychain)

macos_keychain_lock(keychain = NULL)

macos_keychain_unlock(keychain = NULL, password = NULL)

macos_keychain_is_locked(keychain = NULL)

macos_item_attr()

macos_item_match_options()
}
\arguments{
\item{value}{Value of the item, a password, key or certificate. It must
a raw vector or a string. If it is a string, then it is converted to
UTF-8.}

\item{attributes}{Narrow the search by indicating the attributes that
the found item or items should have.}

\item{class}{Type of items to search, see \code{\link[=macos_item_classes]{macos_item_classes()}} for
possible values.}

\item{item}{Keychain item, creted via \code{\link[=macos_item]{macos_item()}} or returned
by oskeyking itself.}

\item{keychain}{Keychain to use. \code{NULL} means the default one.}

\item{match}{Condition the search in a variety of ways. For example, you
can limit the results to a specific number of items, control case
sensitivity when matching string attributes, etc. See 'Search parameters'
below.}

\item{return_data}{Whether to include the secret data in the
search result. If this is set to \code{TRUE}, then you'll have to set the
\code{limit} parameter (in the \code{match} argument) to a finite value.
If this is \code{TRUE}, then macOS will prompt you for passwords if necessary.
You might get multiple password prompts, if you set \code{limit} to a larger
than one value.}

\item{update}{Named list specifying the new values of attributes.}

\item{password}{Password to unlock the keychain, or new password to
set when creating a new keychain. May be \code{NULL} in interactive
sessions, to force a secure password dialog.}

\item{domain}{The preference domain from which you wish to retrieve
the keychain search list:
\itemize{
\item \code{"all"}: include all keychains currently on the search list,
\item \code{"user"}: user preference domain,
\item \code{"system"}: system or daemon preference domain,
\item \code{"common"}: keychains common to everyone,
\item \code{"dynamic"}: dynamic search list (typically provided by removable
keychains such as smart cards).
}}
}
\value{
\code{macos_item_classes()} returns a character vector, the
names of the supported keychain item classes.

\code{macos_item()} returns a new \code{oskeyring_macos_item}
object.

\code{macos_item_add()} returns \code{NULL}, invisibly.

\code{macos_item_search()} returns a list of keychain items.

\code{macos_item_update()} returns \code{NULL}, invisibly.

\code{macos_item_delete()} returns \code{NULL}, invisibly.

\code{macos_keychain_create()} returns \code{NULL}, invisibly.

\code{macos_keychain_list()} returns a data frame with columns:
\itemize{
\item \code{path}: Path to the file of the keychain.
\item \code{is_locked}: Whether the keychain is locked.
\item \code{is_readable}: Whether the keychain is readable by the user.
\item \code{is_writeable}: Whether the keychain is writeable by the user.
}

\code{macos_keychain_delete()} returns \code{NULL}, invisibly.

\code{macos_keychain_lock()} returns \code{NULL}, invisibly.

\code{macos_keychain_unlock()} returns \code{NULL}, invisibly.

\code{macos_keychain_is_locked()} returns \code{TRUE} or \code{FALSE}.

\code{macos_item_attr()} returns a list of lists of character
scalars, the description of keychain item attributes, for each
keychain item class.

\code{macos_item_match_options()} returns a list of character
scalars, the description of the supported match options.
}
\description{
\verb{macos_item_*} functions add, delete, update and search Keychain items.

\verb{macos_keychain_*} functions create, delete, list, lock, unlock
keychains.

\code{macos_item_classes()} lists the supported Keychain item classes.
\code{macos_item_attr()} lists the supported attributes for these classes.
\code{macos_item_match_options()} lists the options supported by the
\code{match} argument of \code{macos_item_search()}.
}
\section{Keychain items}{
\code{macos_item_classes()} returns the currently supported Keychain item
classes.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_item_classes()
#> [1] "generic_password"  "internet_password"
}\if{html}{\out{</div>}}

\code{macos_item()} creates a new Keychain item. See the next section about
the attributes that are supported for the various item types.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{it <- macos_item("secret", list(service = "My service", account = "Gabor"))
it
#> <oskeyring_macos_item: generic_password>
#>  account: Gabor
#>  service: My service
#>  value: <-- hidden -->
}\if{html}{\out{</div>}}

\code{macos_item_add()} adds an item to the keychain. If there is already an
item with the same primary keys, then it will error.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_item_add(it)
}\if{html}{\out{</div>}}

\code{macos_item_search()} searches for Keychain items. If \code{return_data} is
\code{TRUE} then it also returns the secret data. Returning the secret data
might create a password entry dialog. If \code{return_data} is \code{TRUE} then
you need to set the \code{limit} match condition to a (small) finite number.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_item_search(attributes = list(service = "My service"))
#> [[1]]
#> <oskeyring_macos_item: generic_password>
#>  account: Gabor
#>  creation_date: 2023-11-03 12:30:13
#>  label: My service
#>  modification_date: 2023-11-03 12:30:13
#>  service: My service
}\if{html}{\out{</div>}}

\code{macos_item_update()} updates existing Keychain items.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_item_update(
  attributes = list(service = "My service", account = "Gabor"),
  update = list(account = "Gabor Csardi")
)
macos_item_search(attributes = list(service = "My service"))
#> [[1]]
#> <oskeyring_macos_item: generic_password>
#>  account: Gabor Csardi
#>  creation_date: 2023-11-03 12:30:13
#>  label: My service
#>  modification_date: 2023-11-03 12:30:13
#>  service: My service
}\if{html}{\out{</div>}}

\code{macos_item_delete()} deletes one or more Keychain items. Note that
all matching items will be deleted.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_item_delete(attributes = list(service = "My service"))
macos_item_search(attributes = list(service = "My service"))
#> list()
}\if{html}{\out{</div>}}
\subsection{Keychain Item Attributes}{
\itemize{
\item The set of supported attributes depends on the class of the item.
\item oskeyring supports the following item classes currently:
generic_password, internet_password.
\item A subset of the attributes form a \emph{primary key}. It is not possible
to add more than one item with the same primary key. See the
primary keys for the various classes below.
\item oskeyring does not currently support all attributes that the
Keychain Services AIP supports.
\item Some attributes are read-only. If you try to set them when adding
or updating items, they will be ignored.
\item If an attribute is not included in the return value of
\code{macos_item_search()} then it is not set, and its default value is in
effect.
}
\subsection{Attributes for generic passwords}{
\itemize{
\item \code{creation_date}: [.POSIXct(1)][read-only] The date the item was created.
\item \code{modification_date}: [.POSIXct(1)][read-only] The last time the item was updated.
\item \code{description}: [character(1)] User-visible string describing this kind ofitem (for example, 'Disk image password').
\item \code{comment}: [character(1)] User-editable comment for this item.
\item \code{label}: [character(1)] User-visible label for this item.
\item \code{is_invisible}: [logical(1)] \code{TRUE} if the item is invisible (that is, should not be displayed).
\item \code{is_negative}: [logical(1)] Indicates whether there is a valid password associated with this keychain item. This is useful if your application doesn't want a password for some particular service to be stored in the keychain, but prefers that it always be entered by the user.
\item \code{account}: [character(1)][key] Account name.
\item \code{service}: [character(1)][key] The service associated with this item.
\item \code{generic}: [character(1)] User-defined attribute.
\item \code{synchronizable}: [logical(1)] Indicates whether the item in question is synchronized to other devices through iCloud.
}
}

\subsection{Attributes for internet passwords}{
\itemize{
\item \code{creation_date}: [.POSIXct(1)][read-only] The date the item was created.
\item \code{modification_date}: [.POSIXct(1)][read-only] The last time the item was updated.
\item \code{description}: [character(1)] User-visible string describing this kind ofitem (for example, 'Disk image password').
\item \code{comment}: [character(1)] User-editable comment for this item.
\item \code{label}: [character(1)] User-visible label for this item.
\item \code{is_invisible}: [logical(1)] \code{TRUE} if the item is invisible (that is, should not be displayed).
\item \code{is_negative}: [logical(1)] Indicates whether there is a valid password associated with this keychain item. This is useful if your application doesn't want a password for some particular service to be stored in the keychain, but prefers that it always be entered by the user.
\item \code{account}: [character(1)][key] Account name.
\item \code{synchronizable}: [logical(1)] Indicates whether the item in question is synchronized to other devices through iCloud.
\item \code{security_domain}: [character(1)][key] The item's security domain.
\item \code{server}: [character(1)][key] Contains the server's domain name or IP address.
\item \code{protocol}: [character(1)][key] The protocol for this item.
\item \code{authentication_type}: character[1][key] Authentication type.
\item \code{port}: [integer(1)][key] Internet port number.
\item \code{path}: [character(1)][key] A path, typically the path component of the URL
}
}

}
}

\section{Search Parameters}{
osxkeychain only supports a limited set of search parameters.
You can provide these for \code{macos_item_search()} as the \code{match} argument:
\itemize{
\item \code{limit}: [numeric(1)] This value specifies the maximum number of results to return or otherwise act upon. Use \code{Inf} to specify all matching items.
}
}

\section{Keychains}{
macOs supports multiple keychains.
There is always a default keychain, which is the user's login keychain,
unless configured differently.
There is also a keychain search list.
Keychains may belong into four non-exclusive categories, see the
\code{domain} argument of \code{macos_keychain_list()}.
A keychain is stored in an encrypted file on the disk, see the first
column of the output of \code{macos_keychain_list()}.

\verb{macos_item_*()} functions have a \code{keychain} argument to direct or
restrict the operation to a single keychain only. These are the defaults:
\itemize{
\item \code{macos_item_add()} adds the item to the default keychain.
\item \code{macos_item_search()} searches all keychains in the search list.
\item \code{macos_item_update()} updates matching items on all keychains in the
search list.
\item \code{macos_item_delete()} deletes matching items from all keychains in the
search list.
}

\code{macos_keychain_create()} creates a new keychain.

\code{macos_keychain_list()} lists all keychains on the search list.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{new <- "~/Library/Keychains/test.keychain-db"
macos_keychain_create(new, password = "secret")
macos_keychain_list()
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{##                                                     path is_unlocked
## 1 /Users/gaborcsardi/Library/Keychains/login.keychain-db        TRUE
## 2 /Users/gaborcsardi/Library/Keychains/shiny.keychain-db       FALSE
## 3  /Users/gaborcsardi/Library/Keychains/test.keychain-db        TRUE
## 4                     /Library/Keychains/System.keychain       FALSE
##   is_readable is_writeable
## 1        TRUE         TRUE
## 2        TRUE        FALSE
## 3        TRUE         TRUE
## 4        TRUE        FALSE
}\if{html}{\out{</div>}}

\code{macos_keychain_lock()} locks a keychain.
\code{macos_keychain_unlock()} unlocks a keychain.
\code{macos_keychain_is_locked()} checks if a keychain is locked.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_keychain_lock(new)
macos_keychain_is_locked(new)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{## [1] TRUE
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_keychain_unlock(new, password = "secret")
macos_keychain_is_locked(new)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{## [1] FALSE
}\if{html}{\out{</div>}}

\code{macos_keychain_delete()} deletes a keychain: it removes it from the
search list and deletes the data from the disk. It currently refuses to
delete the user's login keychain and the system keychain. Use Keychain
Access instead if you want to delete these. (Only do this if you are
aware of the bad consequences.)

\if{html}{\out{<div class="sourceCode r">}}\preformatted{macos_keychain_delete(new)
macos_keychain_list()
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{##                                                     path is_unlocked
## 1 /Users/gaborcsardi/Library/Keychains/login.keychain-db        TRUE
## 2 /Users/gaborcsardi/Library/Keychains/shiny.keychain-db       FALSE
## 3                     /Library/Keychains/System.keychain       FALSE
##   is_readable is_writeable
## 1        TRUE         TRUE
## 2        TRUE        FALSE
## 3        TRUE        FALSE
}\if{html}{\out{</div>}}
}

\examples{
# See above
}
\seealso{
The Keychain Services API documentation at
\url{https://developer.apple.com/documentation/security/keychain_services}.
}
